/*
 * Project: MoleCuilder
 * Description: creates and alters molecular systems
 * Copyright (C)  2010 University of Bonn. All rights reserved.
 * Please see the LICENSE file or "Copyright notice" in builder.cpp for details.
 */
/*
 * ManipulablePrototypeFactoryUnitTest.cpp
 *
 *  Created on: Jan 03, 2011
 *      Author: heber
 */
// include config.h
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <cppunit/CompilerOutputter.h>
#include <cppunit/extensions/TestFactoryRegistry.h>
#include <cppunit/ui/text/TestRunner.h>
#include "CodePatterns/Assert.hpp"

#include "ManipulablePrototypeFactoryUnitTest.hpp"

#include "stubs/CommonStub.hpp"
#include "stubs/ManipulableCloneStub.hpp"
#include "stubs/ManipulablePrototypeFactoryStub.hpp"

#include <typeinfo>

#ifdef HAVE_TESTRUNNER
#include "UnitTestMain.hpp"
#endif /*HAVE_TESTRUNNER*/

/********************************************** Test classes **************************************/

// Registers the fixture into the 'registry'
CPPUNIT_TEST_SUITE_REGISTRATION( ManipulablePrototypeFactoryTest );

void ManipulablePrototypeFactoryTest::setUp()
{
  rndA =
      ManipulablePrototypeFactoryStub::getInstance().
      ManipulablePrototypeTable[ManipulablePrototypeFactoryStub::Aclass]->clone();
  CPPUNIT_ASSERT_EQUAL( 0,  rndA->getcount() );
  rndB =
      ManipulablePrototypeFactoryStub::getInstance().
      ManipulablePrototypeTable[ManipulablePrototypeFactoryStub::Bclass]->clone();
  CPPUNIT_ASSERT_EQUAL( 0,  rndB->getcount() );

  rndA_1 = NULL;
  rndA_2 = NULL;
  rndA_3 = NULL;
}

void ManipulablePrototypeFactoryTest::tearDown()
{
  delete rndA;
  delete rndA_1;
  delete rndA_2;
  delete rndA_3;
  delete rndB;
  ManipulablePrototypeFactoryStub::purgeInstance();
}

void ManipulablePrototypeFactoryTest::DistributionTest()
{
  // check the injectiveness of enum and string map
  for (ManipulablePrototypeFactoryStub::NameMap::const_iterator
      iter = ManipulablePrototypeFactoryStub::getInstance().names.begin();
      iter != ManipulablePrototypeFactoryStub::getInstance().names.end();
      ++iter) {
    CPPUNIT_ASSERT_EQUAL(
        iter->second,
        ManipulablePrototypeFactoryStub::getInstance().getName(
            ManipulablePrototypeFactoryStub::getInstance().getEnum(
                iter->second
            )
        )
    );
  }

  // check distributions in the table
  CPPUNIT_ASSERT_EQUAL(
      std::string(typeid(ManipulablePrototype<teststubs::Aclass>).name()),
      std::string(typeid(*rndA).name())
  );
  CPPUNIT_ASSERT_EQUAL(
      std::string(typeid(ManipulablePrototype<teststubs::Bclass>).name()),
      std::string(typeid(*rndB).name())
  );
}

void ManipulablePrototypeFactoryTest::getProductEnumTest()
{
  rndA_1 =
      ManipulablePrototypeFactoryStub::getInstance().getProduct(ManipulablePrototypeFactoryStub::Aclass);
  CPPUNIT_ASSERT( typeid(*rndA) == typeid(*rndA_1) );
}

void ManipulablePrototypeFactoryTest::getProductNameTest()
{
  rndA_1 =
      ManipulablePrototypeFactoryStub::getInstance().getProduct(std::string("Aclass"));
  CPPUNIT_ASSERT( typeid(*rndA) == typeid(*rndA_1) );
}

void ManipulablePrototypeFactoryTest::getProductTypeTest()
{
  rndA_1 =
      ManipulablePrototypeFactoryStub::getInstance().getProduct( typeid(teststubs::Aclass) );
  CPPUNIT_ASSERT( typeid(*rndA) == typeid(*rndA_1) );
}

void ManipulablePrototypeFactoryTest::Individualitytest()
{
  // increasing this does not change the prototype
  rndA->count();

  rndA_1 = ManipulablePrototypeFactoryStub::getInstance().getProduct(ManipulablePrototypeFactoryStub::Aclass);
  CPPUNIT_ASSERT_EQUAL( 0,  rndA_1->getcount() );
  CPPUNIT_ASSERT( rndA->getcount() != rndA_1->getcount() );
  rndA_1->count();
  CPPUNIT_ASSERT( rndA->getcount() == rndA_1->getcount() );

  rndA_2 = ManipulablePrototypeFactoryStub::getInstance().getProduct(ManipulablePrototypeFactoryStub::Aclass);
  CPPUNIT_ASSERT_EQUAL( 0,  rndA_2->getcount() );
  CPPUNIT_ASSERT( rndA->getcount() != rndA_2->getcount() );
  rndA_2->count();
  CPPUNIT_ASSERT( rndA->getcount() == rndA_2->getcount() );
  CPPUNIT_ASSERT( rndA_1->getcount() == rndA_2->getcount() );
}

void ManipulablePrototypeFactoryTest::getPrototypetest()
{
  // this method is protected and only friends may access it.
  IManipulablePrototype &rndA_1c = ManipulablePrototypeFactoryStub::getInstance().getPrototype(std::string("Aclass"));
  CPPUNIT_ASSERT_EQUAL( 0, rndA_1c.getcount() );

  // clone the type and check whether new default values holds
  rndA_2 = ManipulablePrototypeFactoryStub::getInstance().getProduct(ManipulablePrototypeFactoryStub::Aclass);
  CPPUNIT_ASSERT_EQUAL( rndA_1c.getcount(), rndA_2->getcount() );
  rndA_2->count();
  CPPUNIT_ASSERT( rndA_1c.getcount() != rndA_2->getcount() );

  // the following is possible (with PrototypeFactory not!)
  rndA_1c.count();
  CPPUNIT_ASSERT_EQUAL( 1, rndA_1c.getcount() );

  // rndA has been cloned before we have manipulated the prototype
  CPPUNIT_ASSERT( rndA->getcount() != rndA_2->getcount() );

  // prototype is at 1, has been manipulated!
  rndA_3 = ManipulablePrototypeFactoryStub::getInstance().getProduct(ManipulablePrototypeFactoryStub::Aclass);
  CPPUNIT_ASSERT_EQUAL( 1, rndA_3->getcount() );

  // rndA is older, before manipulation, hence still at 0
  CPPUNIT_ASSERT( rndA->getcount() != rndA_3->getcount() );
}

void ManipulablePrototypeFactoryTest::manipulatePrototypetest()
{
  teststubs::classParameters params;
  params.counter = 100;
  ManipulablePrototypeFactoryStub::getInstance().manipulatePrototype(std::string("Aclass"), params);

  rndA_1 = ManipulablePrototypeFactoryStub::getInstance().getProduct(std::string("Aclass"));
  CPPUNIT_ASSERT( 100 == rndA_1->getcount() );

  rndA_1->count();
  CPPUNIT_ASSERT( 101 == rndA_1->getcount() );

  rndA_2 = ManipulablePrototypeFactoryStub::getInstance().getProduct(std::string("Aclass"));
  CPPUNIT_ASSERT( 100 == rndA_2->getcount() );
  CPPUNIT_ASSERT( rndA_1->getcount() != rndA_2->getcount() );
}
